#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>

#include "../include/string.h"

#include "obj.h"
#include "sarfio.h"

#include "guiutils.h"
#include "cdialog.h"
#include "csd.h"
#include "editclist.h"

#include "editor.h"
#include "editorcb.h"
#include "editortex.h"
#include "editorobj.h"
#include "editorscnopen.h"
#include "editorlist.h"
#include "scnedit.h"
#include "config.h"


static gint EditorSceneryLoadProgressCB(gpointer data, glong n, glong m);
static void EditorSceneryLoadIterate(
        editor_struct *e, GtkCList *clist, gint *cur_row,
        gpointer p, gint i, gint total,
        GtkWidget *toplevel, gboolean verbose
);
void EditorSceneryLoad(
	editor_struct *e, const gchar *path, gboolean verbose
);
void EditorSceneryUnload(editor_struct *e);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)     (((x) < 0) ? ((x) * -1) : (x))
#define IS_STRING_EMPTY(s)      (((s) != NULL) ? (*(s) == '\0') : TRUE)

#define RADTODEG(r)     ((r) * 180 / PI)
#define DEGTORAD(d)     ((d) * PI / 180)


/*
 *	Load progress callback, this just updates the status bar.
 */
static gint EditorSceneryLoadProgressCB(gpointer data, glong n, glong m)
{
	editor_struct *e = EDITOR(data);
	if((e == NULL) || (m <= 0))
	    return(-1);

	if(EDITOR_STOP_COUNT(e))
	    return(-1);

	StatusBarProgress(
	    e->status_bar, (gfloat)n / (gfloat)m, TRUE
	);

	return(0);
}


/*
 *	Called by EditorSceneryLoad() to load the scenery file parameter
 *	p as an object and add it to the clist.
 */
static void EditorSceneryLoadIterate(
	editor_struct *e, GtkCList *clist, gint *cur_row,
	gpointer p, gint i, gint total,
	GtkWidget *toplevel, gboolean verbose
)
{
	gint type;


#define GET_CURRENT_OBJECT()	\
 ((*cur_row > -1) ? EDITOR_OBJ(gtk_clist_get_row_data(clist, *cur_row)) : NULL)

	if(p == NULL)
	    return;

	e->hdr_freeze_count++;

	/* Get parameter type. */
	type = *(gint *)p;

	/* Version. */
        if(type == SAR_PARM_VERSION)
        {
#if 0
            sar_parm_version_struct *v =
                (sar_parm_version_struct *)p;
#endif
	}
        /* Name. */
        else if(type == SAR_PARM_NAME)
        {
            GtkWidget *w;
            sar_parm_name_struct *v =
                (sar_parm_name_struct *)p;

            w = e->hdr_name_entry;
	    if(w != NULL)
		gtk_entry_set_text(
		    GTK_ENTRY(w), (v->name != NULL) ? v->name : ""
		);
        }
	/* Scene registered location. */
	else if(type == SAR_PARM_REGISTER_LOCATION)
        {
            sar_parm_register_location_struct *v =
                (sar_parm_register_location_struct *)p;
            GtkCList *clist = (GtkCList *)EDITCLIST_CLIST(
		e->hdr_registered_locations_eclist
	    );
            if((clist != NULL) ? (clist->columns >= 7) : FALSE)
            {
                gint row;
                gchar *strv[7];
                strv[0] = STRDUP(v->name);
                strv[1] = g_strdup_printf("%.2f", (gfloat)v->pos.x);
                strv[2] = g_strdup_printf("%.2f", (gfloat)v->pos.y);
                strv[3] = g_strdup_printf("%.2f",
		    (gfloat)SFMMetersToFeet(v->pos.z)
		);
                strv[4] = g_strdup_printf("%.2f",
		    (gfloat)RADTODEG(v->dir.heading)
		);
                strv[5] = g_strdup_printf("%.2f",
                    (gfloat)RADTODEG(v->dir.pitch)
                );
                strv[6] = g_strdup_printf("%.2f",
                    (gfloat)RADTODEG(v->dir.bank)
                );
                row = gtk_clist_append(clist, strv);
                g_free(strv[0]);
                g_free(strv[1]);
                g_free(strv[2]);
                g_free(strv[3]);
                g_free(strv[4]);
                g_free(strv[5]);
                g_free(strv[6]);
	    }
        }
        /* Scene GPS. */
        else if(type == SAR_PARM_SCENE_GPS)
        {
	    GtkAdjustment *adj;
	    GtkWidget *w;
            sar_parm_scene_gps_struct *v =
                (sar_parm_scene_gps_struct *)p;

	    w = e->hdr_gps_longitude_spin;
	    adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
	    if(adj != NULL)
	    {
		adj->value = (gfloat)v->dms_x_offset;
		gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
	    }
            w = e->hdr_gps_latitude_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            if(adj != NULL)
            {
                adj->value = (gfloat)v->dms_y_offset;
                gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
            }
            w = e->hdr_planetary_radius_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            if(adj != NULL)
            {
                adj->value = (gfloat)v->planet_radius;
                gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
            }
        }
        /* Scene map. */
        else if(type == SAR_PARM_SCENE_MAP)
        {
	    GtkAdjustment *adj;
	    GtkWidget *w;
            sar_parm_scene_map_struct *v =
                (sar_parm_scene_map_struct *)p;

            w = e->hdr_map_width_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            if(adj != NULL)
            {
                adj->value = (gfloat)v->width;
                gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
            }

            w = e->hdr_map_height_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            if(adj != NULL)
            {
                adj->value = (gfloat)v->height;
                gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
            }

	    if(v->file != NULL)
	    {
		const gchar *filename = v->file;
		EditorTexDelete(e->hdr_map_tex);
		e->hdr_map_tex = EditorTexNewLoad(
		    filename,
		    NULL,
		    filename,
		    1.0f
		);
	    }
	}
        /* Scene elevation. */
        else if(type == SAR_PARM_SCENE_ELEVATION)
        {
            GtkAdjustment *adj;
            GtkWidget *w;
            sar_parm_scene_elevation_struct *v =
                (sar_parm_scene_elevation_struct *)p;

            w = e->hdr_elevation_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            if(adj != NULL)
            {
                adj->value = (gfloat)SFMMetersToFeet(v->elevation);
                gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
            }
        }
        /* Scene cant. */
        else if(type == SAR_PARM_SCENE_CANT)
        {
            GtkAdjustment *adj;
            GtkWidget *w;
            sar_parm_scene_cant_struct *v =
                (sar_parm_scene_cant_struct *)p;

            w = e->hdr_cant_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            if(adj != NULL)
            {
                adj->value = (gfloat)RADTODEG(v->cant);
                gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
            }
        }
        /* Scene ground flags. */
        else if(type == SAR_PARM_SCENE_GROUND_FLAGS)
        {
            sar_parm_scene_ground_flags_struct *v =
                (sar_parm_scene_ground_flags_struct *)p;
	    GtkToggleButton *tb = (GtkToggleButton *)e->hdr_ground_is_water_tb;
	    if(tb != NULL)
		gtk_toggle_button_set_active(
		    tb,
		    (v->flags & SAR_SCENE_BASE_FLAG_IS_WATER) ? TRUE : FALSE
		);
        }
        /* Scene ground tile. */
        else if(type == SAR_PARM_SCENE_GROUND_TILE)
        {
            GtkAdjustment *adj;
            GtkWidget *w;
            sar_parm_scene_ground_tile_struct *v =
                (sar_parm_scene_ground_tile_struct *)p;

            w = e->hdr_ground_tile_width_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            if(adj != NULL)
            {
                adj->value = (gfloat)v->tile_width;
                gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
            }
            w = e->hdr_ground_tile_height_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            if(adj != NULL)
            {
                adj->value = (gfloat)v->tile_height;
                gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
            }

            w = e->hdr_ground_tile_close_range_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            if(adj != NULL)
            {
                adj->value = (gfloat)v->close_range;
                gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
            }

	    w = e->hdr_ground_color_btn;
	    if(w != NULL)
	    {
		csd_color_struct c;
		c.a = (gfloat)v->color.a;
		c.r = (gfloat)v->color.r;
		c.g = (gfloat)v->color.g;
		c.b = (gfloat)v->color.b;
		CSDColorButtonSetColor(w, &c);
	    }

	    w = e->hdr_ground_texture_entry;
	    if(w != NULL)
		gtk_entry_set_text(
		    GTK_ENTRY(w), (v->texture_name != NULL) ?
			v->texture_name : ""
		);
        }
        /* Texture base directory. */
        else if(type == SAR_PARM_TEXTURE_BASE_DIRECTORY)
        {
#if 0
            sar_parm_texture_base_directory_struct *v =
                (sar_parm_texture_base_directory_struct *)p;
#endif
        }
        /* Texture load. */
        else if(type == SAR_PARM_TEXTURE_LOAD)
        {
	    GtkCList *clist = (GtkCList *)EDITCLIST_CLIST(e->hdr_textures_eclist);
            sar_parm_texture_load_struct *v =
                (sar_parm_texture_load_struct *)p;
	    if((clist != NULL) ? (clist->columns >= 3) : FALSE)
	    {
		gint row;
		gchar *strv[3];
		strv[0] = STRDUP(v->name);
		strv[1] = STRDUP(v->file);
		strv[2] = g_strdup_printf("%.2f", (gfloat)v->priority);
		row = gtk_clist_append(clist, strv);
		g_free(strv[0]);
                g_free(strv[1]);
                g_free(strv[2]);
	    }
        }

	/* Create a new object. */
	else if(type == SAR_PARM_NEW_OBJECT)
	{
	    sar_parm_new_object_struct *v =
		(sar_parm_new_object_struct *)p;
	    editor_obj_struct *obj = EditorObjNew();
	    obj->type = v->object_type;
	    *cur_row = EditorObjListInsertRow(e, -1,  obj);
	}
	/* Create a new helipad object. */
	else if(type == SAR_PARM_NEW_HELIPAD)
	{
            sar_parm_new_helipad_struct *v =
                (sar_parm_new_helipad_struct *)p;
            editor_obj_struct *obj = EditorObjNew();
            obj->type = SAR_OBJ_TYPE_HELIPAD;
	    g_free(obj->ref_obj_name);
	    obj->ref_obj_name = STRDUP(v->ref_obj_name);
	    obj->ref_obj_num = -1;
	    memcpy(&obj->ref_offset, &v->ref_offset, sizeof(sar_position_struct));
            memcpy(&obj->ref_dir, &v->ref_dir, sizeof(sar_direction_struct));

	    obj->helipad_flags = v->flags;
	    g_free(obj->helipad_type_name);
	    obj->helipad_type_name = STRDUP(v->style);
	    obj->helipad_length = (gfloat)v->length;
	    obj->helipad_width = (gfloat)v->width;
	    obj->helipad_recession = (gfloat)v->recession;
	    g_free(obj->helipad_label);
	    obj->helipad_label = STRDUP(v->label);

            *cur_row = EditorObjListInsertRow(e, -1,  obj);
	}
	/* Create a new runway object. */
        else if(type == SAR_PARM_NEW_RUNWAY)
        {
            sar_parm_new_runway_struct *v =
                (sar_parm_new_runway_struct *)p;
            editor_obj_struct *obj = EditorObjNew();
            obj->type = SAR_OBJ_TYPE_RUNWAY;
	    obj->range = (gfloat)v->range;

	    obj->runway_flags = v->flags;
	    obj->runway_length = (gfloat)v->length;
	    obj->runway_width = (gfloat)v->width;
            obj->runway_surface = v->surface;

            obj->runway_dashes = v->dashes;
            obj->runway_edge_light_spacing =
		(gfloat)v->edge_light_spacing;

	    g_free(obj->runway_north_label);
	    obj->runway_north_label = STRDUP(v->north_label);
	    g_free(obj->runway_south_label);
	    obj->runway_south_label = STRDUP(v->south_label);

            obj->runway_north_displaced_threshold =
		(gfloat)v->north_displaced_threshold;
            obj->runway_south_displaced_threshold =
		(gfloat)v->south_displaced_threshold;

            *cur_row = EditorObjListInsertRow(e, -1,  obj);
        }
        /* Create a new human object. */
        else if(type == SAR_PARM_NEW_HUMAN)
        {
            sar_parm_new_human_struct *v =
                (sar_parm_new_human_struct *)p;
            editor_obj_struct *obj = EditorObjNew();
            obj->type = SAR_OBJ_TYPE_HUMAN;

	    obj->human_flags = v->flags;
	    g_free(obj->human_type_name);
	    obj->human_type_name = STRDUP(v->type_name);

            *cur_row = EditorObjListInsertRow(e, -1,  obj);
        }
        /* Create a new fire object. */
        else if(type == SAR_PARM_NEW_FIRE)
        {
            sar_parm_new_fire_struct *v =
                (sar_parm_new_fire_struct *)p;
            editor_obj_struct *obj = EditorObjNew();
            obj->type = SAR_OBJ_TYPE_FIRE;

	    obj->fire_radius = (gfloat)v->radius;
	    obj->fire_height = (gfloat)v->height;

            *cur_row = EditorObjListInsertRow(e, -1,  obj);
        }
        /* Create a new smoke object. */
        else if(type == SAR_PARM_NEW_SMOKE)
        {
            sar_parm_new_smoke_struct *v =
                (sar_parm_new_smoke_struct *)p;
            editor_obj_struct *obj = EditorObjNew();
            obj->type = SAR_OBJ_TYPE_SMOKE;

	    memcpy(&obj->smoke_offset, &v->offset, sizeof(sar_position_struct));
	    obj->smoke_radius_start = (gfloat)v->radius_start;
	    obj->smoke_radius_max = (gfloat)v->radius_max;
	    obj->smoke_radius_rate = (gfloat)v->radius_rate;
	    obj->smoke_hide_at_max = v->hide_at_max ? TRUE : FALSE;
	    obj->smoke_respawn_int = v->respawn_int;
	    obj->smoke_total_units = v->total_units;
	    obj->smoke_color_code = v->color_code;

            *cur_row = EditorObjListInsertRow(e, -1,  obj);
        }
        /* Create a new premodeled object. */
        else if(type == SAR_PARM_NEW_PREMODELED)
        {
            sar_parm_new_premodeled_struct *v =
                (sar_parm_new_premodeled_struct *)p;
            editor_obj_struct *obj = EditorObjNew();
            obj->type = SAR_OBJ_TYPE_PREMODELED;

	    g_free(obj->premodeled_type_name);
	    obj->premodeled_type_name = STRDUP(v->model_type);
            g_free(obj->premodeled_value);
	    obj->premodeled_value = NULL;
	    for(i = 0; i < v->argc; i++)
	    {
		obj->premodeled_value = strcatalloc(obj->premodeled_value, v->argv[i]);
		if(i < (v->argc - 1))
		    obj->premodeled_value = strcatalloc(obj->premodeled_value, " ");
	    }

            *cur_row = EditorObjListInsertRow(e, -1,  obj);
        }
        /* Select object by name. */
        else if(type == SAR_PARM_SELECT_OBJECT_BY_NAME)
        {
	    sar_parm_select_object_by_name_struct *v =
		(sar_parm_select_object_by_name_struct *)p;
	    if(v->name != NULL)
	    {
		*cur_row = EditorObjListMatchByName(e, v->name);
	    }
	}
        /* Model file. */
        else if(type == SAR_PARM_MODEL_FILE)
        {
	    sar_parm_model_file_struct *v =
		(sar_parm_model_file_struct *)p;
	    editor_obj_struct *obj = GET_CURRENT_OBJECT();
	    if(obj != NULL)
	    {
		g_free(obj->model_file);
		obj->model_file = STRDUP(v->file);
	    }
        }
        /* Range. */
        else if(type == SAR_PARM_RANGE)
        {
            sar_parm_range_struct *v =
                (sar_parm_range_struct *)p;
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
            if(obj != NULL)
            {
		obj->range = (gfloat)v->range;
            }
        }
        /* Range far. */
        else if(type == SAR_PARM_RANGE_FAR)
        {
            sar_parm_range_far_struct *v =
                (sar_parm_range_far_struct *)p;
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
            if(obj != NULL)
            {
                obj->range_far = (gfloat)v->range_far;
            }
        }
        /* Translate. */
        else if(type == SAR_PARM_TRANSLATE)
        {
            sar_parm_translate_struct *v =
                (sar_parm_translate_struct *)p;
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
            if(obj != NULL)
            {
		memcpy(&obj->pos, &v->translate, sizeof(sar_position_struct));
            }
        }
        /* Translate random. */
        else if(type == SAR_PARM_TRANSLATE_RANDOM)
        {
#if 0
            sar_parm_translate_random_struct *v =
                (sar_parm_translate_random_struct *)p;
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
            if(obj != NULL)
            {

            }
#endif
        }
        /* Rotate. */
        else if(type == SAR_PARM_ROTATE)
        {
            sar_parm_rotate_struct *v =
                (sar_parm_rotate_struct *)p;
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
            if(obj != NULL)
            {
                memcpy(&obj->dir, &v->rotate, sizeof(sar_direction_struct));
            }
        }
        /* No depth test. */
        else if(type == SAR_PARM_NO_DEPTH_TEST)
        {
#if 0
            sar_parm_no_depth_test_struct *v =
                (sar_parm_no_depth_test_struct *)p;
#endif
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
            if(obj != NULL)
            {
                obj->flags |= SAR_OBJ_FLAG_NO_DEPTH_TEST;
            }
        }
        /* Polygon offset. */
        else if(type == SAR_PARM_POLYGON_OFFSET)
        {
            sar_parm_polygon_offset_struct *v =
                (sar_parm_polygon_offset_struct *)p;
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
            if(obj != NULL)
            {
		obj->flags |= v->flags;
	    }
        }
        /* Contact bounds spherical. */
        else if(type == SAR_PARM_CONTACT_BOUNDS_SPHERICAL)
        {
            sar_parm_contact_bounds_spherical_struct *v =
                (sar_parm_contact_bounds_spherical_struct *)p;
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
            if(obj != NULL)
            {
                sar_contact_bounds_struct *cb = &obj->cb;
		cb->crash_flags |= SAR_CRASH_FLAG_CRASH_CAUSE;
                cb->crash_type = SAR_CRASH_TYPE_OBSTRUCTION;
                cb->contact_shape = SAR_CONTACT_SHAPE_SPHERICAL;
                cb->contact_radius = v->radius;
            }
        }
        /* Contact bounds cylendical. */
        else if(type == SAR_PARM_CONTACT_BOUNDS_CYLENDRICAL)
        {
            sar_parm_contact_bounds_cylendrical_struct *v =
                (sar_parm_contact_bounds_cylendrical_struct *)p;
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
            if(obj != NULL)
            {
                sar_contact_bounds_struct *cb = &obj->cb;
                cb->crash_flags |= SAR_CRASH_FLAG_CRASH_CAUSE |
		    SAR_CRASH_FLAG_SUPPORT_SURFACE;
		cb->crash_type = SAR_CRASH_TYPE_OBSTRUCTION;
                cb->contact_shape = SAR_CONTACT_SHAPE_CYLENDRICAL;
                cb->contact_radius = v->radius;
                cb->contact_h_min = v->height_min;
                cb->contact_h_max = v->height_max;
            }
        }
        /* Contact bounds rectangular. */
        else if(type == SAR_PARM_CONTACT_BOUNDS_RECTANGULAR)
        {
            sar_parm_contact_bounds_rectangular_struct *v =
                (sar_parm_contact_bounds_rectangular_struct *)p;
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
            if(obj != NULL)
            {
                sar_contact_bounds_struct *cb = &obj->cb;
                cb->crash_flags |= SAR_CRASH_FLAG_CRASH_CAUSE |
		    SAR_CRASH_FLAG_SUPPORT_SURFACE;
                cb->crash_type = SAR_CRASH_TYPE_OBSTRUCTION;
                cb->contact_shape = SAR_CONTACT_SHAPE_RECTANGULAR;
                cb->contact_x_min = v->x_min;
                cb->contact_x_max = v->x_max;
                cb->contact_y_min = v->y_min;
                cb->contact_y_max = v->y_max;
                cb->contact_z_min = v->z_min;
                cb->contact_z_max = v->z_max;
            }
        }
        /* Ground elevation. */
        else if(type == SAR_PARM_GROUND_ELEVATION)
        {
#if 0
            sar_parm_ground_elevation_struct *v =
                (sar_parm_ground_elevation_struct *)p;
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
            if(obj != NULL)
            {

            }
#endif
        }
        /* Name. */
        else if(type == SAR_PARM_OBJECT_NAME)
        {
            sar_parm_object_name_struct *v =
                (sar_parm_object_name_struct *)p;
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
            if(obj != NULL)
            {
		g_free(obj->name);
		obj->name = STRDUP(v->name);
		EditorObjListUpdateRow(e, *cur_row, obj);
            }
        }
        /* Map description. */
        else if(type == SAR_PARM_OBJECT_MAP_DESCRIPTION)
        {
            sar_parm_object_map_description_struct *v =
                (sar_parm_object_map_description_struct *)p;
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
            if(obj != NULL)
            {
                g_free(obj->description);
                obj->description = STRDUP(v->description);
            }
        }
        /* Fuel. */
        else if(type == SAR_PARM_FUEL)
        {
#if 0
            sar_parm_fuel_struct *v =
                (sar_parm_fuel_struct *)p;
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
            if(obj != NULL)
            {

            }
#endif
        }
        /* Hit points. */
        else if(type == SAR_PARM_HITPOINTS)
        {
#if 0
            sar_parm_hitpoints_struct *v =
                (sar_parm_hitpoints_struct *)p;
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
            if(obj != NULL)
            {

            }
#endif
        }
        /* Engine state. */
        else if(type == SAR_PARM_ENGINE_STATE)
        {
#if 0
            sar_parm_engine_state_struct *v =
                (sar_parm_engine_state_struct *)p;
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
            if(obj != NULL)
            {

            }
#endif
        }
        /* Passengers. */
        else if(type == SAR_PARM_PASSENGERS)
        {
#if 0
            sar_parm_passengers_struct *v =
                (sar_parm_passengers_struct *)p;
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
            if(obj != NULL)
            {

            }
#endif
        }
        /* Runway approach lighting north. */
        else if(type == SAR_PARM_RUNWAY_APPROACH_LIGHTING_NORTH)
        {
            sar_parm_runway_approach_lighting_north_struct *v =
                (sar_parm_runway_approach_lighting_north_struct *)p;
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
	    if((obj != NULL) ? (obj->type == SAR_OBJ_TYPE_RUNWAY) : FALSE)
            {
		obj->runway_north_lighting_flags = v->flags;
	    }
	}
        /* Runway approach lighting south. */
        else if(type == SAR_PARM_RUNWAY_APPROACH_LIGHTING_SOUTH)
        {
            sar_parm_runway_approach_lighting_south_struct *v =
                (sar_parm_runway_approach_lighting_south_struct *)p;
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
            if((obj != NULL) ? (obj->type == SAR_OBJ_TYPE_RUNWAY) : FALSE)
            {
                obj->runway_south_lighting_flags = v->flags;
            }
        }
        /* Human message enter. */
        else if(type == SAR_PARM_HUMAN_MESSAGE_ENTER)
        {
            sar_parm_human_message_enter_struct *v =
                (sar_parm_human_message_enter_struct *)p;
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
            if((obj != NULL) ? (obj->type == SAR_OBJ_TYPE_HUMAN) : FALSE)
            {
		g_free(obj->human_message_enter);
		obj->human_message_enter = STRDUP(v->message);
            }
        }
        /* Human reference. */
        else if(type == SAR_PARM_HUMAN_REFERENCE)
        {
            sar_parm_human_reference_struct *v =
                (sar_parm_human_reference_struct *)p;
            editor_obj_struct *obj = GET_CURRENT_OBJECT();
            if((obj != NULL) ? (obj->type == SAR_OBJ_TYPE_HUMAN) : FALSE)
            {
		g_free(obj->ref_obj_name);
		obj->ref_obj_name = STRDUP(v->reference_name);
                obj->ref_obj_num = -1;
		obj->human_flags |= v->flags;
            }
        }

        e->hdr_freeze_count--;

	EditorSceneryLoadProgressCB(e, i + 1, total);

#undef GET_CURRENT_OBJECT
}

/*
 *	Load scenery to editor.
 */
void EditorSceneryLoad(editor_struct *e, const gchar *path, gboolean verbose)
{
	gpointer *parm = NULL;
	gint i, cur_row = -1, total_parms = 0;
	GtkWidget *toplevel;
	GtkCList *clist;


#define STATUS_BAR_MESSAGE(m)	\
 StatusBarMessage(e->status_bar, (m), TRUE);


	if(e == NULL)
	    return;

        if(EDITOR_PROCESSING(e))
            return;

	/* Unload existing scenery first, resetting the has_changes
	 * mark and all widget values.
	 */
	EditorSceneryUnload(e);

	/* If no path is given then do not open anything. */
	if((path != NULL) ? (*path == '\0') : TRUE)
	    return;

        toplevel = EDITOR_TOPLEVEL(e);
	clist = (GtkCList *)e->obj_clist;
	if(clist == NULL)
	    return;

	EditorStopReset(e, TRUE);
        e->processing = TRUE;

	if(verbose)
	    STATUS_BAR_MESSAGE("Reading scenery file...");

	/* Load scenery file into parameters. */
	if(SARParmLoadFromFile(
	    path, SAR_FILE_FORMAT_SCENE,
	    &parm, &total_parms,
	    -1,
	    e,
	    EditorSceneryLoadProgressCB
	))
	{
	    if(verbose)
	    {
		gchar *buf = g_strdup_printf(
		    "Error opening scenery file:\n\n    %s",
		    path
		);
                CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
		    "Open Error",
		    buf,
"The scenery file may be corrupt or the specified scenery file\n\
does not exist.",
		    CDIALOG_ICON_ERROR,
		    CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
		    CDIALOG_BTNFLAG_OK
		);
		CDialogSetTransientFor(NULL);
		g_free(buf);
		STATUS_BAR_MESSAGE("Error opening scenery file");
		StatusBarProgress(e->status_bar, 0.0f, TRUE);
	    }
            e->processing = FALSE;
	    EditorStopReset(e, FALSE);
	    return;
	}

	/* Scenery file parameters have now been loaded, now begin
	 * parsing the parameters as objects and add them to the
	 * objects list.
	 */
        if(verbose)
	{
            STATUS_BAR_MESSAGE("Parsing scenery data...");
            StatusBarProgress(e->status_bar, 0.0f, TRUE);
	}

	/* Update scenery file name on editor. */
	g_free(e->scn_filename);
	e->scn_filename = g_strdup(path);

	gtk_clist_freeze(clist);

	/* Iterate through all parameters and load objects. */
	for(i = 0; i < total_parms; i++)
	{
	    if(EDITOR_STOP_COUNT(e))
		break;

	    EditorSceneryLoadIterate(
		e, clist, &cur_row,
		parm[i], i, total_parms,
		toplevel, verbose
	    );
	}

	/* Realize all loaded objects. */
	EditorObjListRealizeAll(e, TRUE);

	gtk_clist_thaw(clist);

	/* Delete all parameters. */
	SARParmDeleteAll(&parm, &total_parms);

	if(verbose)
	{
	    STATUS_BAR_MESSAGE("Scenery file loaded");
            StatusBarProgress(e->status_bar, 0.0f, TRUE);
	}

	e->processing = FALSE;
	EditorStopReset(e, FALSE);

#undef STATUS_BAR_MESSAGE
}

/*
 *	Unload scenery on editor.
 */
void EditorSceneryUnload(editor_struct *e)
{
	gchar **s;
	GtkWidget *w;
	GtkCList *clist;


	if(e == NULL)
	    return;

	if(EDITOR_PROCESSING(e))
	    return;

	e->processing = TRUE;

	/* Delete all objects. */
	clist = (GtkCList *)e->obj_clist;
	if(clist != NULL)
	{
	    gtk_clist_freeze(clist);
	    gtk_clist_clear(clist);
	    gtk_clist_thaw(clist);
	}

#define FREE_STRING	\
{ if(*s != NULL) { g_free(*s); *s = NULL; } }
#define TB_RESET	\
{ if(w != NULL) { gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE); }}
#define ENTRY_RESET	\
{ if(w != NULL) { gtk_entry_set_text(GTK_ENTRY(w), ""); } }
#define SPIN_RESET						\
{ if(w != NULL) {						\
 GtkAdjustment *adj = GTK_SPIN_BUTTON(w)->adjustment;		\
 if(adj != NULL) {						\
  adj->value = (gfloat)MAX(0.0, adj->lower);			\
  gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");	\
 }								\
} }
#define CLIST_RESET						\
{ if(clist != NULL) {						\
 gtk_clist_freeze(clist);					\
 gtk_clist_clear(clist);					\
 gtk_clist_thaw(clist);						\
} }

	/* Begin reseting values. */
	e->has_changes = FALSE;
	s = &e->scn_filename;
	FREE_STRING
	s = &e->scn_title;
        FREE_STRING
        s = &e->scn_description;
        FREE_STRING

	/* Header values. */
        e->hdr_freeze_count++;
	w = e->hdr_name_entry;
	ENTRY_RESET
	w = e->hdr_gps_longitude_spin;
	SPIN_RESET
        w = e->hdr_gps_latitude_spin;
        SPIN_RESET
        w = e->hdr_planetary_radius_spin;
        SPIN_RESET
        w = e->hdr_cant_spin;
        SPIN_RESET
        w = e->hdr_elevation_spin;
        SPIN_RESET
	w = e->hdr_ground_is_water_tb;
	TB_RESET
        w = e->hdr_ground_tile_width_spin;
	SPIN_RESET
        w = e->hdr_ground_tile_height_spin;
	SPIN_RESET
        w = e->hdr_ground_tile_close_range_spin;
	SPIN_RESET
/*	w = e->hdr_ground_color_btn; */
	w = e->hdr_ground_texture_entry;
	ENTRY_RESET
	w = e->hdr_map_width_spin;
        SPIN_RESET
	w = e->hdr_map_height_spin;
	SPIN_RESET
	clist = (GtkCList *)EDITCLIST_CLIST(e->hdr_textures_eclist);
	CLIST_RESET
        clist = (GtkCList *)EDITCLIST_CLIST(e->hdr_registered_locations_eclist);
        CLIST_RESET

	EditorTexDelete(e->hdr_map_tex);
	e->hdr_map_tex = NULL;

        e->hdr_freeze_count--;


#undef CLIST_RESET
#undef SPIN_RESET
#undef ENTRY_RESET
#undef TB_RESET
#undef FREE_STRING

        e->processing = FALSE;
}
